
import asyncio


try:
    from qtpy import QtWidgets
    has_qt_error = False
        
except Exception as e:
    has_qt_error = e
    QtWidgets = None


try:
    from qtpy.QtWebEngineWidgets import QWebEngineView
    has_webview_error = False
        
except Exception as e:
    has_webview_error = e
    QWebEngineView = None

try:
    from qasync import QSelectorEventLoop
    has_qasync_error = False
    has_qasync = True
        
except Exception as e:
    has_qasync = False
    has_qasync_error = e
    QSelectorEventLoop = None



class GraphicalBackend:


    def __init__(self, parent=None) -> None:
        from herre.auth import get_current_herre
        self.herre = get_current_herre()
        self.parent = parent
        self.spawned_app = None


    def __enter__(self):
        global has_qt_error
        assert not has_qt_error, f"You cannot run with a Qt Backend if no QT Backend is installed. Please install PyQT5  or PySide {str(has_qt_error)}"

        if QtWidgets.QApplication.instance() is None:
            # if it does not exist then a QApplication is created
            self.spawned_app = QtWidgets.QApplication([])

        return self


    def __exit__(self, *args, **kwargs):
        if self.spawned_app: self.spawned_app.exit()


    async def __aenter__(self):
        assert QtWidgets.QApplication.instance() is not None, "You cannot call this Outside of an with_qt wrapped Application"
        assert isinstance(self.herre.loop, QSelectorEventLoop), "You cannot assign outside of an with_qt wrapped APplication"
        return self.__enter__()

    async def __aexit__(self, type, value, traceback):
        
        if type is not None:
            if issubclass(type, asyncio.CancelledError): 
                print("Raising cancellation")
                raise type(value).with_traceback(traceback)

            if issubclass(type, Exception):
                print(f"Raising exceütopm {type} {value} {traceback}")
                raise type(value).with_traceback(traceback)


            raise type(value).with_traceback(traceback)


from abc import ABC, abstractmethod

class QueryVariable:

    @abstractmethod
    async def to_variable(self):
        raise NotImplementedError



async def parse_variables(variables):

    parsed_kwargs = {}

    for key, value in variables.items():
        if isinstance(value, QueryVariable):
            parsed_kwargs[key] = await value.to_variable()
        else:
            parsed_kwargs[key] = value


    return parsed_kwargs


from abc import abstractclassmethod, abstractmethod
from typing import Generic, Optional, Type, TypeVar
from pydantic import BaseModel
from pydantic.main import ModelMetaclass
from herre.wards.base import BaseWard


ModelType = TypeVar("ModelType", bound="Model", covariant=True)
ClassType = TypeVar("ClassType", covariant=True)

class BaseModelManager:

    def __init__(self, modelClass: ModelType = None) -> None:
        self.modelClass =  modelClass
        super().__init__()

    def __call__(self, modelClass: ModelType):
        self.modelClass = modelClass
        return self

    @property
    def ward(self) -> BaseWard:
        from herre.wards.registry import get_ward_registry
        return get_ward_registry().get_ward_instance(self.modelClass.Meta.ward)



class SyncModelManager(Generic[ModelType], BaseModelManager):
    """ A Manager to get Models from Packer
    syncrhonisouly

    Args:
        Generic ([type]): [description]

    Returns:
        [type]: [description]
    """

    @abstractmethod
    def get(self, id: str, **kwargs) -> ModelType:
        """Fetches an instance of Model on 

        Args:
            id (str): The unque UI

        Returns:
            ModelType: The Model
        """
        raise NotImplementedError("Please use a Subclass")


class AsyncModelManager(Generic[ModelType], BaseModelManager):


    @abstractmethod
    async def get(self, id: str = None, **kwargs) -> ModelType:
        """Fetches an instance of Model on 

        Args:
            id (str, optional): The unque UI

        Returns:
            ModelType: The Model
        """
        raise NotImplementedError("Please use a Subclass")

    
    @abstractmethod
    async def create(self, **kwargs) -> ModelType:
        """Create an instance of Model on the Website 

        Args:
            id (str, optional): The unque UI

        Returns:
            ModelType: The Model
        """
        raise NotImplementedError("Please use a Subclass")


class ModelMeta(ModelMetaclass):
    """
    The Model Meta class extends the Pydantic Metaclass and adds in 
    syncrhonous and asynchronous Managers. These Managers allow direct
    interaction with serverside Objects mimicking the Django ORM Scheme
    (https://docs.djangoproject.com/en/3.2/topics/db/queries/) it
    also registeres the Model as a potential serializer.

    Every Class using this metaclass has to subclass pydantic.BaseModel and
    implement a Meta class with the identifier attribute set to a cleartext
    identifier on the arkitekt platform.

    If this identifier does not exist, the serializer can potentially be auto
    registered with the platform according to the apps name

    Args:
        ModelMetaclass ([type]): [description]
    """


    def __new__(mcls, name, bases, attrs):
        slots = set(attrs.pop('__slots__', tuple())) # The slots from: https://github.com/samuelcolvin/pydantic/issues/655#issuecomment-610900376
        for base in bases:
            if hasattr(base, '__slots__'):
                slots.update(base.__slots__)

        if '__dict__' in slots:
            slots.remove('__dict__')
        attrs['__slots__'] = tuple(slots)

        mcls.overriden_manager = attrs.pop("objects") if "objects" in attrs else None
        mcls.overriden_async_manager = attrs.pop("asyncs") if "asyncs" in attrs else None
        return super(ModelMeta, mcls).__new__(mcls, name, bases, attrs)

    @property
    def objects(cls: Type[ModelType]) -> SyncModelManager[ModelType]:
        if not cls.__objects:
            cls.__objects = cls.get_objectsclass()(cls)
        return cls.__objects

    @property
    def asyncs(cls: Type[ModelType]) -> AsyncModelManager[ModelType]:
        if not cls.__asyncs:
            cls.__asyncs = cls.get_asyncclass()(cls)
        return cls.__asyncs


    def __init__(self, name, bases, attrs):
        super(ModelMeta, self).__init__(name, bases, attrs)
        if attrs["__qualname__"] != "Model":
            # This gets allso called for our Baseclass which is abstract
            self.__meta = attrs["Meta"] if "Meta" in attrs else None
            assert self.__meta is not None, f"Please provide a Meta class in your Arnheim Model {name}"

            try:
                if self.__meta.abstract:
                    return
            except:
                pass

            self.__objects = self.overriden_manager(self) if self.overriden_manager else None
            self.__asyncs =  self.overriden_async_manager(self) if self.overriden_async_manager else None
            
            register = getattr(self.__meta, "register", True)
            assert hasattr(self.__meta, "ward"), f"Please specifiy which Ward this Model should use in Meta of  {attrs['__qualname__']}"

            if register:
                self.register_model(meta=self.__meta)



class Model(BaseModel, metaclass=ModelMeta):
    """Model

    Model is the abstract baseclass of all Serverside Models and provides a Django ORM
    like interface for retrieving data from the Server.

    Implements:
        id (str): Every Model has an id (UUID) that identifies the Server Instance

    Args:
        BaseModel ([type]): [description]
        metaclass ([type], optional): [description]. Defaults to ModelMeta.

    Raises:
        NotImplementedError: [description]
        NotImplementedError: [description]

    Returns:
        [type]: [description]
    """
    id: Optional[str]

    @abstractclassmethod
    def get_objectsclass(cls):
        raise NotImplementedError("Please provide an Implementatoin of SyncModelManager")

    @abstractclassmethod
    def get_asyncclass(cls):
        raise NotImplementedError("Please provide an Implementatoin of SyncModelManager")

    @classmethod
    def register_model(cls):
        """ Will be invoked by the meta class and will make this model
        register with other models if wanted

        Returns:
            [type]: [description]
        """
        return None





from abc import abstractmethod
from herre.grants.models import User

from pydantic.main import prepare_config
from herre.console.context import get_current_console
from herre.config.herre import HerreConfig
from abc import ABC
import os 
import asyncio
import requests
import logging
import aiohttp


logger = logging.getLogger(__name__)

class BaseGrant(ABC):
    refreshable = False

    def __init__(self, config: HerreConfig, token_url="o/token/", authorize_url="o/authorize/") -> None:
        self.config = config
        if not config.secure: logger.warn("Using Insecure Oauth2 Protocol.. Please only for local and debug deployments")
        os.environ["OAUTHLIB_INSECURE_TRANSPORT"] = "0" if self.config.secure else "1"


        self.base_url = f'{"https" if self.config.secure else "http"}://{self.config.host}:{self.config.port}/'
        self.auth_url = self.base_url + authorize_url
        self.token_url = self.base_url + token_url
        self.userinfo_url = self.base_url + "userinfo/"
        self.refresh_url = self.token_url
        self._userinfo = None
       
        self.scopes = self.config.scopes + ["introspection"]
        self.scope = " ".join(self.scopes)
        self.token = None
        self._user = None
        # State changed through Class
        self.can_refresh = self.refreshable


    @abstractmethod
    def fetchToken(self):
        raise NotImplementedError()

    @property
    def user(self) -> User:
        return self._user


    @property
    def logged_in(self):
        return self.token is not None

    @property
    def access_token(self):
        assert self.token is not None, "You need to login before you can use this function. Call Herre.login()"
        return self.token["access_token"]


    def login(self, **kwargs):
        """IS SYNC!!!!
        """
        with get_current_console().status("[bold green] Authenticating"):
            self.fetchToken()
            assert "access_token" in self.token, "Returned token does not have an access_token"
            if "refresh_token" in self.token:
                self.can_refresh = True

        try:
            self._user = User(**requests.get(self.userinfo_url, headers={"Authorization": f"Bearer {self.access_token}"}).json())
        except:
            self._user = None


    async def refresh(self, **kwargs):
        assert "refresh_token" in self.token, "Token had not refresh-token attached"
        async with aiohttp.ClientSession() as session:
            async with session.post(self.refresh_url, data={"grant_type": "refresh_token", "refresh_token": self.token["refresh_token"], "client_id": self.config.client_id, "client_secret": self.config.client_secret}) as resp:
                self.token =  await resp.json()
                assert "access_token" in self.token, "Returned refreshed token does not have an access_token"
                if "refresh_token" in self.token:
                    self.can_refresh = True




from herre.config.base import BaseConfig
from pydantic import Field
from enum import Enum
from typing import List, Optional


class GrantType(str, Enum):
    IMPLICIT = "IMPLICIT"
    PASSWORD = "PASSWORD"
    CLIENT_CREDENTIALS = "CLIENT_CREDENTIALS"
    AUTHORIZATION_CODE = "AUTHORIZATION_CODE"
    AUTHORIZATION_CODE_SERVER = "AUTHORIZATION_CODE_SERVER"

class HerreConfig(BaseConfig):

    secure: bool 
    host: str
    port: int
    client_id: str 
    client_secret: str
    authorization_grant_type: GrantType
    scopes: List[str]
    redirect_uri: Optional[str]
    jupyter_sync: bool = False

    class Config:
        yaml_group = "herre"
        env_prefix = "herre_"


    def __str__(self) -> str:
        return f"{'Secure' if self.secure else 'Insecure'} Connection to {self.host}:{self.port} on Grant {self.authorization_grant_type}"

    






from herre.grants.code_server.app import AuthorizationCodeServerGrant
from herre.wrap_qt import wrap_qt_in_loop
from herre.logging import setLogging
from herre.jupy import in_notebook
from herre.grants.code.app import AuthorizationCodeGrant
from herre.grants.backend.app import BackendGrant
from herre.grants.backend.app import BackendGrant
from herre.config.herre import GrantType, HerreConfig
import asyncio 
from threading import Thread
import os
import logging
from herre.loop import loopify
import time

logger = logging.getLogger(__name__)

class HerreError(Exception):
    pass

def newloop(loop):
    asyncio.set_event_loop(loop)
    try:
        loop.run_forever()
    finally:
        print("RUnning Here")
        loop.run_until_complete(loop.shutdown_asyncgens())
        loop.close()


class HerreClient:

    def __init__(self,
        config_path = "bergen.yaml",
        username: str = None,
        password: str = None,
        allow_insecure = False,
        force_async = False,
        force_sync = False,
        register = True,
        qt = None,
        auto_login=True,
        **overrides,
    ) -> None:
        """ Creates A Herre Client

        Args:
            config_path (str, optional): [description]. Defaults to "bergen.yaml".
            username (str, optional): [description]. Defaults to None.
            password (str, optional): [description]. Defaults to None.
            allow_insecure (bool, optional): [description]. Defaults to False.
            in_sync (bool, optional): Should we force an in_sync modus if an event loop is already running. Loop will be send to another thread. Defaults to True.

        Raises:
            HerreError: [description]
        """

        self.loop = None
        self.thread_id = None
        self.config_path = config_path

        setLogging(config_path=config_path)

        self.config = HerreConfig.from_file(config_path, **overrides)

        try:
            loop = asyncio.get_running_loop()
            if loop.is_running():
                logger.info("We are in an already running async loop!")
                if in_notebook():
                    logger.info("We are running in a notebook! Choosing on saved preference!")
                    if self.config.jupyter_sync:
                        thread_loop = asyncio.new_event_loop()
                        self.thread = Thread(target=newloop, args=(thread_loop,))
                        self.thread.start()
                        logger.info("Running in Seperate Thread so that we can use the sync syntax")
                        self.loop = thread_loop
                    else:
                        self.loop = loop
                else:
                    logger.info("We are now running in an async Context")
                    self.loop = loop
            else:
                if qt is not None:
                    logger.info("Wrapping Qt app so that we can call functions asynchronously")
                    self.loop = wrap_qt_in_loop(qt)

                self.loop = loop

        except RuntimeError as e:
            logger.info("There is no event-loop in this thread. Lets create a new One")
            if force_sync:
                thread_loop = asyncio.new_event_loop()
                self.thread = Thread(target=newloop, args=(thread_loop,))
                self.thread.start()
                logger.info("Running in Seperate Thread so that we can use the sync syntax")
                self.loop = thread_loop
            else:
                loop = asyncio.new_event_loop()
                asyncio.set_event_loop(loop)
                self.loop = loop

        if self.config.authorization_grant_type == GrantType.CLIENT_CREDENTIALS:
            self.grant = BackendGrant(self.config)

        elif self.config.authorization_grant_type == GrantType.AUTHORIZATION_CODE:
            self.grant = AuthorizationCodeGrant(self.config)

        elif self.config.authorization_grant_type == GrantType.AUTHORIZATION_CODE_SERVER:
            self.grant = AuthorizationCodeServerGrant(self.config)

        else:
            raise HerreError("Unknown GrantType")

        if register:
            set_current_herre(self)

        if auto_login:
            self.login(username=username, password=password)


    def login(self, username: str = None, password: str = None):
        self.grant.login(username=username, password=password)


    def close(self):
        self.loop.call_soon_threadsafe(self.loop.stop())

        while self.loop.is_running():
            print("Waiting to kill")
            time.sleep(0.1)


    
    async def refresh(self):
        assert self.grant.can_refresh, "Grant is not able to refresh. Please login again"
        await self.grant.refresh()


    @property
    def logged_in(self):
        return self.grant.logged_in

    @property
    def user(self):
        assert self.grant.logged_in, "User is not logged in"
        return self.grant.user

    @property
    def headers(self, **additionals):
        assert self.grant.access_token is not None, "Access token is not set yet, please login?"
        return {"Authorization": f"Bearer {self.grant.access_token}", **additionals}




CURRENT_HERRE = None

def get_current_herre(**kwargs):
    global CURRENT_HERRE
    if not CURRENT_HERRE:
        CURRENT_HERRE = HerreClient(**kwargs)
    return CURRENT_HERRE

def set_current_herre(herre):
    global CURRENT_HERRE
    CURRENT_HERRE = herre


from herre.graphical import GraphicalBackend, has_webview_error
from herre.console.context import console, get_current_console
from herre.grants.base import BaseGrant
from oauthlib.oauth2 import WebApplicationClient
from requests_oauthlib.oauth2_session import OAuth2Session
from aiohttp import ClientSession
import requests
import webbrowser
from http.server import BaseHTTPRequestHandler, HTTPServer






class RedirectHandler(BaseHTTPRequestHandler):
    backend = None
    stop_server = False
    token = False



    def do_GET(self):
        self.send_response(200)
        self.send_header("Content-type", "text/html")
        self.end_headers()
        print(self.path)
        RedirectHandler.token = self.backend.session.fetch_token(self.backend.token_url, client_secret=self.backend.config.client_secret, authorization_response=self.path)
        self.wfile.write(bytes("<html><head><title>https://pythonbasics.org</title></head>", "utf-8"))
        self.wfile.write(bytes("<p>Request: %s</p>" % self.path, "utf-8"))
        self.wfile.write(bytes('<body>', "utf-8"))
        self.wfile.write(bytes('''<script>
                                setTimeout(window.close(), 5000);
                                </script>''', "utf-8"))
        self.wfile.write(bytes("<p>You can close this website now!</p>", "utf-8"))
        self.wfile.write(bytes("</body></html>", "utf-8"))
        RedirectHandler.stop_server = True
     
        




class AuthorizationCodeServerGrant(BaseGrant):
    refreshable = True

    def fetchToken(self, **kwargs):

        self.web_app_client = WebApplicationClient(self.config.client_id, scope=self.scope)

        # Create an OAuth2 session for the OSF
        self.session = OAuth2Session(
            self.config.client_id, 
            self.web_app_client,
            scope=self.scope, 
            redirect_uri=self.config.redirect_uri,
            
        )

        auth_url, state = self.session.authorization_url(self.auth_url)
        webbrowser.open(auth_url)

        RedirectHandler.backend = self

        httpd = HTTPServer(("localhost", 6767), RedirectHandler)
        print("Server started http://%s:%s" % ("localhost", 6767))

        while not RedirectHandler.stop_server:
            httpd.handle_request()
        
        self.token = RedirectHandler.token
        return RedirectHandler.token




from qtpy import QtCore
from qtpy.QtWidgets import QDialog, QVBoxLayout
from qtpy.QtWebEngineWidgets import QWebEngineView, QWebEnginePage
import logging
import os

logger = logging.getLogger(__name__)

class WebEnginePage(QWebEnginePage):

    def javaScriptConsoleMessage(self, level, msg, line, sourceID):
        pass

class LoginWindow(QWebEngineView):
    """ A Login window for the OSF """

    def __init__(self, backend = None, tokenCallback=None):
        super(LoginWindow, self).__init__()
        self.state = None
        self.session = backend.session
        self.backend = backend

        auth_url, state = self.session.authorization_url(self.backend.auth_url)

        self.urlChanged.connect(self.check_URL)
        self.callback = tokenCallback
        self.setPage(WebEnginePage(self))
        self.load(QtCore.QUrl(auth_url))
        self.set_state(state)


    def set_state(self,state):
        self.state = state

    def check_URL(self, url: QtCore.QUrl):
        url = url.url()
        if url.startswith(self.backend.config.redirect_uri):
            try:
                token = self.session.fetch_token(self.backend.token_url, client_secret=self.backend.config.client_secret, authorization_response=url)
                if token: self.callback(token)
            except Exception as e:
                logger.exception(e)




class LoginDialog(QDialog):
    def __init__(self, backend = None, parent = None):
        super(LoginDialog, self).__init__(parent)
        self.setModal(True)
        layout = QVBoxLayout(self)

        # nice widget for editing the date
        self.webview = LoginWindow(backend=backend, tokenCallback=self.tokenReady)
        layout.addWidget(self.webview)

        self.token = None
        # OK and Cancel buttons

    def tokenReady(self, token):
        self.token = token
        self.webview.close() # NEEDS TO BE CALLED GOD DAMN IT
        self.accept()

    # static method to create the dialog and return (date, time, accepted)
    @staticmethod
    def getToken(backend = None, parent = None):
        dialog = LoginDialog(backend=backend, parent=parent)
        result = dialog.exec_()
        token = dialog.token
        return token, result == QDialog.Accepted